/*
 @file   CommonFFT.cpp
 @brief  FFT֘ACu֐̎
 @param  AdIn Research,Inc
 @date   2007/02/15
*/

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "CommonFFT.h"
typedef unsigned long DWORD;
#define WDC_PI          ((double) 3.14159265358979323846)


/*----------------------- [eBeB֐` -----------------------*/
static int bitRev( int nBit, int nExp );
static int normDataNum( int nData );
static bool wf_Hanning( int nData, const double* pdIn, double* pdOut );
static bool wf_Hamming( int nData, const double* pdIn, double* pdOut );
static bool wf_Blackman( int nData, const double* pdIn, double* pdOut );
static bool wf_BlackmanHarris( int nData, const double* pdIn, double* pdOut );
static bool wf_Bartlett( int nData, const double* pdIn, double* pdOut );
static bool wf_Parzen( int nData, const double* pdIn, double* pdOut );
static bool wf_Welch( int nData, const double* pdIn, double* pdOut );
static bool wf_Riesz( int nData, const double* pdIn, double* pdOut );
static bool wf_Sin( int nData, const double* pdIn, double* pdOut );
static bool wf_Triangular( int nData, const double* pdIn, double* pdOut );






/*
 @brief      [pXtB^[ (butterworth 2nd order) s
 @param[in]  nData f[^z̗vf
 @param[in]  pdIn  f[^z
 @param[out] pdOut tB^[̃f[^z
 @param[in]  dCutOffFreq [pXtB^[̃JbgItg
 @param[in]  dSampleFreq TvOg
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool LPF_Butterworth2( int nData, const double* pdIn, double* pdOut, double dCutOffFreq, double dSampleFreq )
{
  // check status and arguments
  if ( nData < 3 || pdIn == NULL || pdOut == NULL || dCutOffFreq <= 0.0 || dSampleFreq <= 0.0)
    return false;

  // determine parameter
  double dOmega  = WDC_PI * 2.0 * dCutOffFreq;
  double dDeltaT = 1.0 / dSampleFreq;
  double dPa1    = -(-8.0 + 2.0 * pow(dOmega * dDeltaT, 2.0)) / (4.0 + 2.0 * sqrt(2.0) * dDeltaT * dOmega + pow(dOmega * dDeltaT, 2.0));
  double dPa2    = -(4.0 - 2.0 * sqrt(2.0) * dOmega * dDeltaT + pow(dOmega * dDeltaT, 2.0)) / (4.0 + 2.0 * sqrt(2.0) * dOmega * dDeltaT + pow(dOmega * dDeltaT, 2.0));
  double dPb1    = (pow(dOmega * dDeltaT, 2.0)) / (4.0 + 2.0 * sqrt(2.0) * dDeltaT * dOmega + pow(dOmega * dDeltaT, 2.0));
  double dPb2    = dPb1 * 2.0;
  double dPb3    = dPb1;
  //
  double adOutQue[2] = { 0.0, 0.0 };
  //
  int iOut;
  for (iOut = 0; iOut < 2; iOut++)
    pdOut[iOut] = 0.0;
  //
  for (int i = 0; i < nData - 2; i++) {
    // fold
    double dFold  = (dPa1 * adOutQue[1] + dPa2 * adOutQue[0]) + (dPb1 * pdIn[i + 2] + dPb2 * pdIn[i + 1] + dPb3 * pdIn[i]);
    pdOut[iOut++] = (i >= 1) ? dFold : 0.0;
    //
    adOutQue[0] = adOutQue[1];
    adOutQue[1] = dFold;
  }
  return true;
}

/*
 @brief      ֐Kp
 @param[in]  iWF   ֐̎
 @param[in]  nData f[^z̗vf
 @param[in]  pdIn  f[^z
 @param[out] pdOut ֐Kp̃f[^z
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool Windowing( int iWF, int nData, const double* pdIn, double* pdOut )
{
  // check status
  if (nData < 2 || nData != normDataNum(nData))
    return false;
  // check argument
  if (pdIn == NULL || pdOut == NULL)
    return false;
  // apply window function
  bool bRes;
  switch (iWF) {
  case WLWF_HANNING:
    bRes = wf_Hanning(nData, pdIn, pdOut);
    break;
  case WLWF_HAMMING:
    bRes = wf_Hamming(nData, pdIn, pdOut);
    break;
  case WLWF_BLACKMAN:
    bRes = wf_Blackman(nData, pdIn, pdOut);
    break;
  case WLWF_BLACKMANHARRIS:
    bRes = wf_BlackmanHarris(nData, pdIn, pdOut);
    break;
  case WLWF_BARTLETT:
    bRes = wf_Bartlett(nData, pdIn, pdOut);
    break;
  case WLWF_PARZEN:
    bRes = wf_Parzen(nData, pdIn, pdOut);
    break;
  case WLWF_WELCH:
    bRes = wf_Welch(nData, pdIn, pdOut);
    break;
  case WLWF_RIESZ:
    bRes = wf_Riesz(nData, pdIn, pdOut);
    break;
  case WLWF_SIN:
    bRes = wf_Sin(nData, pdIn, pdOut);
    break;
  case WLWF_TRIANGULAR:
    bRes = wf_Triangular(nData, pdIn, pdOut);
    break;
  case WLWF_NONE:     // WLWF_SQUARE
    memcpy(pdOut, pdIn, sizeof(double) * nData);
    bRes = true;
    break;
  default:
    bRes = false;
    break;
  }
  return bRes;
}

/*
 @brief      FFTs
 @param[in]  nData   f[^z̗vf
 @param[in]  pdIn    f[^z
 @param[out] pdReal  f[^z
 @param[out] pdImag  f[^z
 @param[out] pdPower p[XyNgf[^z
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool FFT( int nData, const double* pdIn, double* pdReal, double* pdImag, double* pdPower )
{
  // check argument
  if( nData < 1 || ! pdIn || ! pdReal || ! pdImag || ! pdPower )
    return false;

  // Initialize working array
  for( int i = 0; i < nData; i++ ) {
    pdReal[i] = pdIn[i];
    pdImag[i] = 0.0;
  }
  // apply FFT
  int nExp = 0;
  for (int i = nData >> 1; i > 0; i >>= 1)
    nExp++;
  int    nCntD = 1 << nExp;
  double dDeg  = 2.0 * WDC_PI / nCntD;
  //
  int k  = 0;
  int j2 = nCntD;
  int j1 = nExp - 1;
  for (int j = 0; j < nExp; j++) {
    j2 >>= 1;
    do {
      for (int i = 0; i < j2; i++) {
        int    n    = k >> j1;
        int    nBit = bitRev(n, nExp);
        double d    = dDeg * nBit;
        double dCos = cos(d);
        double dSin = sin(d);
        int    k1   = k + j2;
        double dA   = pdReal[k1] * dCos + pdImag[k1] * dSin;
        double dB   = pdImag[k1] * dCos - pdReal[k1] * dSin;
        pdReal[k1]     = pdReal[k] - dA;
        pdImag[k1]     = pdImag[k] - dB;
        pdReal[k]      = pdReal[k] + dA;
        pdImag[k]      = pdImag[k] + dB;
        k++;
      }
      k += j2;
    } while (k < nCntD);
    k = 0;
    --j1;
  }
  for (int i = 0; i < nCntD; i++) {
    int nBit = bitRev(i, nExp);
    if (nBit > i) {
      double d  = pdReal[i];
      pdReal[i]    = pdReal[nBit];
      pdReal[nBit] = d;
      d         = pdImag[i];
      pdImag[i]    = pdImag[nBit];
      pdImag[nBit] = d;
    }
  }
  for (int i = 0; i < nCntD; i++) {
    pdPower[i] = sqrt(pdReal[i] * pdReal[i] + pdImag[i] * pdImag[i]) * 2.0 / nCntD;
  }

  return true;
}




/*----------------------- [eBeB֐ -----------------------*/

/*
 @brief      reverse bit order
 @param[in]  nBit input
 @param[in]  nExp shift order
 @return     int
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
int bitRev( int nBit, int nExp )
{
  int nRev = 0;
  for (int i = 0; i < nExp; i++) {
    nRev <<= 1;
    nRev  |= (nBit & 1);
    nBit >>= 1;
  }
  return nRev;
}

/*
 @brief      normalize number to 2^n
 @param[in]  nData 
 @return     int
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
int normDataNum( int nData )
{
  // check argument
  if (nData < 1)
    return 0;

  // find power of two less-or-equals to # of data
  DWORD dwData = (DWORD) nData;
  DWORD dwMask = 0x80000000;
  while (dwMask > 0) {
    if (dwData & dwMask)
      break;
    dwMask >>= 1;
  }
  int nNorm = (int) dwMask;
  // expand?
  if( nData > nNorm )
    nNorm <<= 1;
  return nNorm;
}

/*
 @brief      window function (Hanning)
 @param[in]  nData f[^z̗vf
 @param[in]  pdIn  f[^z
 @param[out] pdOut ֐̃f[^z
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool wf_Hanning( int nData, const double* pdIn, double* pdOut )
{
  // check status
  if (nData < 2 || nData != normDataNum(nData))
    return false;
  // check argument
  if (pdIn == NULL || pdOut == NULL)
    return false;
  // apply window function
  const double dAlpha = 0.5;
  const double dBeta  = 0.5;
  const double dParam = (2.0 * WDC_PI) / (nData - 1);
  for (int i = 0; i < nData; i++)
    pdOut[i] = pdIn[i] * (dAlpha - dBeta * cos(dParam * i));
  return true;
}

/*
 @brief      window function (Hamming)
 @param[in]  nData f[^z̗vf
 @param[in]  pdIn  f[^z
 @param[out] pdOut ֐̃f[^z
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool wf_Hamming( int nData, const double* pdIn, double* pdOut )
{
  // check status
  if (nData < 2 || nData != normDataNum(nData))
    return false;
  // check argument
  if (pdIn == NULL || pdOut == NULL)
    return false;
  // apply window function
  const double dAlpha = 25.0 / 46.0;
  const double dBeta  = 1.0 - dAlpha;
  const double dParam = (2.0 * WDC_PI) / (nData - 1);
  for (int i = 0; i < nData; i++)
    pdOut[i] = pdIn[i] * (dAlpha - dBeta * cos(dParam * i));
  return true;
}

/*
 @brief      window function (Blackman)
 @param[in]  nData f[^z̗vf
 @param[in]  pdIn  f[^z
 @param[out] pdOut ֐̃f[^z
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool wf_Blackman( int nData, const double* pdIn, double* pdOut )
{
  // check status
  if (nData < 2 || nData != normDataNum(nData))
    return false;
  // check argument
  if (pdIn == NULL || pdOut == NULL)
    return false;
  // apply window function
  const double dAlpha = 0.423;
  const double dBeta  = 0.498;
  const double dGamma = 0.0792;
  const double adParam[] = {
    (2.0 * WDC_PI) / (nData - 1),
    (4.0 * WDC_PI) / (nData - 1),
  };
  for (int i = 0; i < nData; i++)
    pdOut[i] = pdIn[i] * (dAlpha - dBeta * cos(adParam[0] * i) + dGamma * cos(adParam[1] * i));
  return true;
}

/*
 @brief      window function (Blackman-Harris)
 @param[in]  nData f[^z̗vf
 @param[in]  pdIn  f[^z
 @param[out] pdOut ֐̃f[^z
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool wf_BlackmanHarris( int nData, const double* pdIn, double* pdOut )
{
  // check status
  if (nData < 2 || nData != normDataNum(nData))
    return false;
  // check argument
  if (pdIn == NULL || pdOut == NULL)
    return false;
  // apply window function
  const double dAlpha = 0.35875;
  const double dBeta  = 0.48829;
  const double dGamma = 0.14128;
  const double dDelta = 0.01168;
  const double adParam[] = {
    (2.0 * WDC_PI) / (nData - 1),
    (4.0 * WDC_PI) / (nData - 1),
    (6.0 * WDC_PI) / (nData - 1),
  };
  for (int i = 0; i < nData; i++)
    pdOut[i] = pdIn[i] * (dAlpha - dBeta * cos(adParam[0] * i) + dGamma * cos(adParam[1] * i) - dDelta * cos(adParam[2] * i));
  return true;
}

/*
 @brief      window function (Bartlett)
 @param[in]  nData f[^z̗vf
 @param[in]  pdIn  f[^z
 @param[out] pdOut ֐̃f[^z
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool wf_Bartlett( int nData, const double* pdIn, double* pdOut )
{
  // check status
  if (nData < 2 || nData != normDataNum(nData))
    return false;
  // check argument
  if (pdIn == NULL || pdOut == NULL)
    return false;
  // apply window function
  const double dParam = 2.0 / (nData - 1);
  double     d;
  for (int i = 0; i < nData; i++) {
    if ((d = i * dParam) > 1.0)
      d = 2.0 - d;
    pdOut[i] = pdIn[i] * d;
  }
  return true;
}

/*
 @brief      window function (Parzen)
 @param[in]  nData f[^z̗vf
 @param[in]  pdIn  f[^z
 @param[out] pdOut ֐̃f[^z
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool wf_Parzen( int nData, const double* pdIn, double* pdOut )
{
  // check status
  if (nData < 2 || nData != normDataNum(nData))
    return false;
  // check argument
  if (pdIn == NULL || pdOut == NULL)
    return false;
  // apply window function
  const double dAlpha = 0.5 * (nData - 1);
  const double dBeta  = dAlpha + 1.0;
  double d;
  for (int i = 0; i < nData; i++) {
    if ((d = (i - dAlpha) / dBeta) > 0.0)
      d = 1.0 - d;
    else
      d = 1.0 + d;
    pdOut[i] = pdIn[i] * d;
  }
  return true;
}

/*
 @brief      window function (Welch)
 @param[in]  nData f[^z̗vf
 @param[in]  pdIn  f[^z
 @param[out] pdOut ֐̃f[^z
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool wf_Welch( int nData, const double* pdIn, double* pdOut )
{
  // check status
  if (nData < 2 || nData != normDataNum(nData))
    return false;
  // check argument
  if (pdIn == NULL || pdOut == NULL)
    return false;
  // apply window function
  const double dAlpha = 0.5 * (nData - 1);
  const double dBeta  = dAlpha + 1.0;
  for (int i = 0; i < nData; i++) {
    double d = (i - dAlpha) / dBeta;
    pdOut[i] = pdIn[i] * (1.0 - d * d);
  }
  return true;
}

/*
 @brief      window function (Riesz)
 @param[in]  nData f[^z̗vf
 @param[in]  pdIn  f[^z
 @param[out] pdOut ֐̃f[^z
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool wf_Riesz( int nData, const double* pdIn, double* pdOut )
{
  // check status
  if (nData < 2 || nData != normDataNum(nData))
    return false;
  // check argument
  if (pdIn == NULL || pdOut == NULL)
    return false;
  // apply window function
  assert(0);
  return false;
}

/*
 @brief      window function (Sin)
 @param[in]  nData f[^z̗vf
 @param[in]  pdIn  f[^z
 @param[out] pdOut ֐̃f[^z
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool wf_Sin( int nData, const double* pdIn, double*  pdOut )
{
  // check status
  if (nData < 2 || nData != normDataNum(nData))
    return false;
  // check argument
  if (pdIn == NULL || pdOut == NULL)
    return false;
  // apply window function
  const double dParam = WDC_PI / (nData - 1);
  for (int i = 0; i < nData; i++)
    pdOut[i] = pdIn[i] * sin(dParam * i);
  return true;
}

/*
 @brief      window function (Triangular)
 @param[in]  nData f[^z̗vf
 @param[in]  pdIn  f[^z
 @param[out] pdOut ֐̃f[^z
 @return     bool
 @param      AdIn Research,Inc
 @date       2007/02/15
*/
bool wf_Triangular( int nData, const double* pdIn, double*  pdOut )
{
  // check status
  if (nData < 2 || nData != normDataNum(nData))
    return false;
  // check argument
  if (pdIn == NULL || pdOut == NULL)
    return false;
  // apply window function
  const double dAlpha = 0.5 * (nData - 1);
  for (int i = 0; i < nData; i++)
    pdOut[i] = pdIn[i] * (1.0 - fabs((i - dAlpha) / dAlpha));
  return true;
}

